(*
    This file is part of the ErrorHandling project - an attempt
    to provide a general framework for error handling in
    Standard ML.

    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)


(*
require "exceptions_sig";
require "$.basis.__text_io.sml";
*)

structure Excn : EXCN =
  struct
    datatype t =
      Generic
    | Bug
    | NotImplemented

    exception EGeneric of (t * string * string) list

    fun eRaise value =
          raise EGeneric([value])

    fun eRaiseBug value =
          eRaise (Bug, "", value)

    fun eRaiseGeneric (valueType, value) =
          eRaise (Generic, valueType, value)

    fun eRaiseNotImplemented value =
          eRaise (NotImplemented, "", value)

    fun reRaise (EGeneric(values)) newValue =
          raise EGeneric(newValue::values)
      | reRaise _ newValue =
          raise EGeneric(newValue::[(Bug, "",
            "Excn.reRaise used with an exception which is not of type EGeneric!")])

    fun outputValues f nil =
          ()
      | outputValues f ((eType, eTypeStr, eMsg)::values) =
          (case eType of
             Generic => f ("Generic exception of type " ^ eTypeStr ^ " raised: ")
           | Bug => f "Software bug: "
           | NotImplemented => f "Feature has not yet been implemented: ";
           f eMsg; f "\n")

    fun output f (EGeneric(values)) =
          outputValues f values
      | output f _ =
          f "Software bug: Excn.output used with an exception which is not of type EGeneric!"

    fun checkedExecOutF outF f arg failValue =
          (f arg)
             handle
               (e as EGeneric(_)) =>
                 (output outF e;
                  failValue)
             (* FIXME: Define the behaviour of checedExec better! *) 
             | e => (outF
                       ("Software bug: Unanticipated exception " ^
                        "of type: " ^ (exnName e) ^ " raised "^
                        "with the following message: " ^ (exnMessage e));
                     failValue)

    fun checkedExec f arg failValue =
          checkedExecOutF
            (fn str => TextIO.output (TextIO.stdErr, str))
            f arg failValue
  end
